# Licensed under a 3-clause BSD style license - see LICENSE.rst

import sep
import numpy as np
import warnings

from astropy.modeling.fitting import LevMarLSQFitter
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.table import Table
from scipy.optimize import curve_fit
from scipy.ndimage.filters import convolve
from sklearn.cluster import DBSCAN

from ._utils import _sep_fix_byte_order
from ..math.moffat import moffat_r, moffat_fwhm, PSFMoffat2D
from ..math.gaussian import gaussian_r, gaussian_fwhm, PSFGaussian2D
from ..math.array import trim_array, xy2r
from ..logger import logger


_default_sharp = (0.2, 1.0)
_default_round = (-1.0, 1.0)


def gen_filter_kernel(size):
    """Generate sextractor like filter kernels."""
    if size == 3:
        return np.array([[0, 1, 0],
                         [1, 4, 1],
                         [0, 1, 0]])
    if size == 5:
        return np.array([[1, 2, 3, 2, 1],
                         [2, 4, 6, 4, 2],
                         [3, 6, 9, 6, 3],
                         [2, 4, 6, 4, 2],
                         [1, 2, 3, 2, 1]])
    if size == 7:
        return np.array([[1, 2, 3, 4, 3, 2, 1],
                         [2, 4, 6, 8, 6, 4, 2],
                         [3, 6, 9, 12, 9, 6, 3],
                         [4, 8, 12, 16, 12, 8, 4],
                         [3, 6, 9, 12, 9, 6, 3],
                         [2, 4, 6, 8, 6, 4, 2],
                         [1, 2, 3, 4, 3, 2, 1]])
    return


def background(data, box_size=64, filter_size=3, mask=None,
               global_bkg=True):
    """Estimate the image background using SExtractor algorithm.

    Parameters
    ----------
    data: array_like
        2D array containing the image to extract the background.
    box_size: `int` (optional)
        Size of background boxes in pixels.
        Default: 64
    filter_size: `int` (optional)
        Filter size in boxes unit.
        Default: 3
    mask: array_like (optional)
        Boolean mask where 1 pixels are masked out in the background
        calculation.
        Default: `None`
    global_bkg: `bool`
        If True, the algorithm returns a single value for background
        and rms, else, a 2D image with local values will be returned.
    """
    d = _sep_fix_byte_order(data)
    if mask is not None:
        mask = np.array(mask, dtype=bool)

    bkg = sep.Background(d, bw=box_size, bh=box_size,
                         fw=filter_size, fh=filter_size,
                         mask=mask)

    if global_bkg:
        return bkg.globalback, bkg.globalrms
    return bkg.back(), bkg.rms()


def sepfind(data, threshold, background, noise,
            mask=None, filter_kernel=3,
            **sep_kwargs):
    """Find sources using SExtractor segmentation algorithm.

    Parameters
    ----------
    data: array_like
        2D array containing the image to extract the source.
    threshold: `float`
        Minimum signal noise desired.
    background: `float`
        Background estimation.
    noise: `float`
        Root-mean-square at each point.
    mask: array_like (optional)
        Boolean mask where 1 pixels are masked out in the background
        calculation.
        Default: `None`
    filter_kernel: `int` (optional)
        Combined filter, which can provide optimal signal-to-noise
        detection for objects with some known shape
        Default: `3`

    sep_kwargs can be any kwargs to be passed to sep.extract function.
    """
    d = _sep_fix_byte_order(data)
    if mask is not None:
        mask = np.array(mask, dtype=bool)

    # Check if need to create a new kernel
    if np.isscalar(filter_kernel):
        filter_kernel = gen_filter_kernel(filter_kernel)

    sep_kwargs['filter_kernel'] = filter_kernel

    sources = sep.extract(d-background, threshold, err=noise, mask=mask,
                          **sep_kwargs)

    if sep_kwargs.get('segmentation_map', False):
        sources = sources[0]  # ignore smap

    return Table(sources)


class DAOFind:
    """Use DAOFind method to detect punctual sources.

    Parameters
    ----------
    fwhm: `int` or `float`
        Default gaussian FWHM for convolution kernel
    sharp_limit: array_like or `None` (optional)
        Low and high cutoff for the sharpness statistic.
        `None` will disable the sharpness filtering.
        Default: (0.2, 1.0)
    round_limit: array_like or `None` (optional)
        Low and high cutoff for the roundness statistic.
        `None` will disable the roundness filtering.
        Default: (-1.0,1.0)
    min_sep: `float` or `None` (optional)
        Minimum separation between sources. None implies it will be
        1.5*sigma.
        Default: None
    skip_invalid_centroid: `bool`
        If `True`, the code will skip the invalid centroid verification and
        include sources with wrong centroid in the final list.
        Default: `False`

    Notes
    -----
    - Adapted from IDL Astro package by D. Jones. Original function available
      at PythonPhot package. https://github.com/djones1040/PythonPhot
      The function recieved some improvements to work better.
    - For our roundness statistics, we use themaximum value between the
      symmetry based roundness (`SROUND` in IRAF DAOFIND) and the marginal
      gaussian fit roundness (`GROUND` in IRAF DAOFIND), allowing better
      identification of assymetric sources in diagonal.
    - Sources are grouped and merged using the DBSCAN algorithm. The minimum
      distance between them is equal to FWHM.
    """

    def __init__(self, fwhm, sharp_limit=_default_sharp,
                 round_limit=_default_round,
                 skip_invalid_centroid=False):
        if fwhm < 0.5:
            raise ValueError('Supplied FWHM must be at least 0.5 pixels')
        # fhwm and related stuff
        self._fwhm = fwhm
        self._sharp_limit = sharp_limit
        self._round_limit = round_limit
        self._skip_invalid_centroid = skip_invalid_centroid
        self._maxbox = 13

        self._compute_convolution_kernel()
        self._compute_constants()

    def _compute_convolution_kernel(self):
        """Generate the proper gaussian kernel."""
        self._sigma2 = (self._fwhm*gaussian_fwhm_to_sigma)**2
        # radius=1.5*sigma, greater then 2
        self._radius = np.max([0.637*self._fwhm, 2.001])
        # index of the center of convolution box
        self._nhalf = np.min([int(self._radius),
                              int((self._maxbox-1)/2.)])
        # size of convolution box. Automatic less or equal maxbox
        self._nbox = 2*self._nhalf + 1
        shape = (self._nbox, self._nbox)

        # valid pixels in convolution kernel, Stetson 'mask'
        self._conv_mask = np.zeros(shape, dtype='int8')

        y, x = np.indices(shape)
        rsq = (x-self._nhalf)**2 + (y-self._nhalf)**2
        self._good = np.where(rsq <= self._radius**2)
        pixels = len(self._good[0])
        self._conv_mask[self._good] = 1
        # convolution kernel, Stetson 'g'
        self._g = np.exp(-0.5*rsq/self._sigma2)
        self._kernel = self._g*self._conv_mask

        # Normalize the convolution kernel and zero not good pixels
        sumc = np.sum(self._kernel)
        sumcsq = np.sum(self._kernel**2) - sumc**2/pixels
        sumc = sumc/pixels
        self._kernel[self._good] = (self._kernel[self._good] - sumc)/sumcsq

        logger.debug('Relative error computed from FWHM: %f',
                     np.sqrt(np.sum(self._kernel[self._good]**2)))

    def _compute_constants(self):
        """Compute some constants needed for statistics."""
        _, x = np.indices((self._nbox, self._nbox))

        self._wt = self._nhalf - np.abs(np.arange(self._nbox)-self._nhalf) + 1
        self._vec = self._nhalf - np.arange(self._nbox)
        self._p = np.sum(self._wt)

        self._c1 = (np.arange(self._nbox)-self._nhalf)**2
        self._c1 = np.exp(-0.5*self._c1/self._sigma2)
        sumc1 = np.sum(self._c1)/self._nbox
        self._c1 = (self._c1-sumc1)/(np.sum(self._c1**2) - sumc1)

        self._xwt = self._nhalf - np.abs(x-self._nhalf) + 1
        self._ywt = np.transpose(self._xwt)

        self._sgx = np.sum(self._g*self._xwt, 1)
        self._sgy = np.sum(self._g*self._ywt, 0)

        self._dgdx = self._sgy*self._vec
        self._dgdy = self._sgx*self._vec

        self._sumgx = np.sum(self._wt*self._sgy)
        self._sumgy = np.sum(self._wt*self._sgx)

        self._sumgsqx = np.sum(self._wt*self._sgx*self._sgx)
        self._sumgsqy = np.sum(self._wt*self._sgy*self._sgy)

        self._sdgdx = np.sum(self._wt*self._dgdx)
        self._sdgdy = np.sum(self._wt*self._dgdy)

        self._sdgdxs = np.sum(self._wt*self._dgdx**2)
        self._sdgdys = np.sum(self._wt*self._dgdy**2)

        self._sgdgdx = np.sum(self._wt*self._sgy*self._dgdx)
        self._sgdgdy = np.sum(self._wt*self._sgx*self._dgdy)

    def _convolve_image(self, image, n_x, n_y):
        """Convolve image with gaussian kernel."""
        h = convolve(image, self._kernel)

        minh = np.min(h)
        nhalf = self._nhalf

        # Fix borders
        h[:, 0:nhalf] = minh
        h[:, n_x-nhalf:n_x] = minh
        h[0:nhalf, :] = minh
        h[n_y-nhalf:n_y - 1, :] = minh

        return h

    def _find_peaks(self, image, index, n_x, n_y):
        """Find peaks above threshold."""
        mask = self._conv_mask.copy()
        mask[self._nhalf, self._nhalf] = 0  # exclude central pixel
        pixels = np.sum(mask)
        good = np.where(mask)
        xx = good[1] - self._nhalf
        yy = good[0] - self._nhalf

        for i in range(pixels):
            hy = index[0]+yy[i]
            hx = index[1]+xx[i]
            # only inside image
            hgood = np.where((hy < n_y) & (hy >= 0) &
                             (hx < n_x) & (hx >= 0))[0]
            stars = np.where(np.greater_equal(image[index[0][hgood],
                                                    index[1][hgood]],
                                              image[hy[hgood],
                                                    hx[hgood]]))
            if len(stars) == 0:
                logger.error('No maxima exceed input threshold.')
                return
            index = np.array([index[0][hgood][stars],
                              index[1][hgood][stars]])

        logger.debug('%i localmaxima located above threshold.',
                     len(index[0]))
        return index

    def _calc_ground(self, chunk):
        """Compute the gaussian marginal fit roundness."""
        dx = np.sum(np.sum(chunk, axis=0)*self._c1)
        dy = np.sum(np.sum(chunk, axis=1)*self._c1)
        if dx <= 0 or dy <= 0:  # invalid roundness
            return np.nan
        return 2*(dx-dy)/(dx+dy)

    def _calc_sround(self, chunk):
        """Compute the folding roundness."""
        # Quads
        # 3 3 4 4 4
        # 3 3 4 4 4
        # 3 3 x 1 1
        # 2 2 2 1 1
        # 2 2 2 1 1
        nhalf = self._nhalf
        chunk = np.array(chunk)
        chunk[nhalf, nhalf] = 0  # copy and setcentral pixel to 0

        q1 = chunk[nhalf:, nhalf+1:].sum()
        q2 = chunk[nhalf+1:, :nhalf+1].sum()
        q3 = chunk[:nhalf+1, :nhalf].sum()
        q4 = chunk[:nhalf, nhalf:].sum()

        sum2 = -q1+q2-q3+q4
        sum4 = q1+q2+q3+q4

        # ignore divide-by-zero RuntimeWarning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            sround = 2.0 * sum2 / sum4
        return sround

    def _calc_sharp(self, chunk, d, mask):
        """Compute the sharpness."""
        center = chunk[self._nhalf, self._nhalf]
        sides = (np.sum(mask*chunk)/np.sum(mask))
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            sharp = (center-sides)/d
        return sharp

    def _calc_centroid(self, chunk, ix, iy):
        """Compute the sources centroid with gaussian fit."""
        # Notes from D.Jones:
        # Modified version in Mar 2008 that differ from original DAOPHOT,
        # which multiplies the correction dx by 1/(1+abs(dx)).
        # DAOPHOT method is more robust (e.g. two different sources will
        # not merge).
        # This method is more accurate and do not introduce biases in the
        # centroid histogram.
        # The change was also applyed in IRAF DAFIND routine.
        # (see http://iraf.net/article.php?story=7211;query=daofind)
        # x centroid
        sd = np.sum(chunk*self._ywt, axis=0)
        sumgd = np.sum(self._wt*self._sgy*sd)
        sumd = np.sum(self._wt*sd)
        sddgdx = np.sum(self._wt*sd*self._dgdx)
        # hx is the height of the best-fitting marginal gaussian
        hx = (sumgd-self._sumgx*sumd/self._p)
        hx /= (self._sumgsqy-self._sumgx**2/self._p)
        if hx <= 0:
            if not self._skip_invalid_centroid:
                return
        skylvl = (sumd - hx*self._sumgx)/self._p
        dx = sddgdx-self._sdgdx*(hx*self._sumgx + skylvl*self._p)
        dx = (self._sgdgdx-dx)/(hx*self._sdgdxs/self._sigma2)
        if np.abs(dx) >= self._nhalf:
            if not self._skip_invalid_centroid:
                return

        # y centroid
        sd = np.sum(chunk*self._xwt, axis=1)
        sumgd = np.sum(self._wt*self._sgx*sd)
        sumd = np.sum(self._wt*sd)
        sddgdy = np.sum(self._wt*sd*self._dgdy)
        hy = (sumgd - self._sumgy*sumd/self._p)
        hy /= (self._sumgsqx - self._sumgy**2/self._p)
        if hy <= 0:
            if not self._skip_invalid_centroid:
                return
        skylvl = (sumd - hy*self._sumgy)/self._p
        dy = sddgdy-self._sdgdy*(hy*self._sumgy + skylvl*self._p)
        dy = (self._sgdgdy-dy)/(hy*self._sdgdys/self._sigma2)
        if np.abs(dy) >= self._nhalf:
            if not self._skip_invalid_centroid:
                return

        return ix+dx, iy+dy

    def _compute(self, image, h, xc, yc):
        """Compute statistics of a single source."""
        nhalf = self._nhalf
        mask = self._conv_mask.copy()
        mask[nhalf, nhalf] = 0  # exclude central pixel

        # original image chunk
        temp = image[yc-nhalf:yc+nhalf+1,
                     xc-nhalf:xc+nhalf+1]
        # convolved image chunk
        temp_conv = h[yc-nhalf:yc+nhalf+1,
                      xc-nhalf:xc+nhalf+1]
        # convolved central pixel intensity
        d = h[yc, xc]

        # compute sharpness
        sh = self._calc_sharp(temp, d, mask)

        # compute roundness
        # maximum value between symmetry and fit roundness
        ground = self._calc_ground(temp)
        sround = self._calc_sround(temp_conv)
        if ground == np.nan:
            return  # invalid source
        rd = ground if abs(ground) >= abs(sround) else sround

        # compute centroid
        centroids = self._calc_centroid(temp, xc, yc)
        xcent = ycent = np.nan
        if centroids is not None:
            xcent, ycent = centroids

        f = d

        return sh, rd, f, xcent, ycent

    def _compute_statistics(self, image, image_convolved, index):
        """Compute source statistics for each star."""
        ngood = len(index[0])
        iy, ix = index

        sources = np.full(ngood, fill_value=np.nan,
                          dtype=[('x', 'f8'), ('y', 'f8'), ('flux', 'f8'),
                                 ('round', 'f4'), ('sharp', 'f4')])

        # loop over all stars
        for i in range(ngood):
            s, r, f, x, y = self._compute(image, image_convolved,
                                          ix[i], iy[i])
            sources[i] = (x, y, f, r, s)

        return sources

    def _filter_sources(self, sources):
        """Perform roundness, sharpness and centroid filtering."""
        # filter stars by sharpness
        sharp_rej = 0
        sharpmask = False
        if self._sharp_limit is not None:
            sharp_limit = sorted(self._sharp_limit)
            sharpmask = np.isnan(sources['sharp'])
            sharpmask |= sources['sharp'] < sharp_limit[0]
            sharpmask |= sources['sharp'] > sharp_limit[1]
            sharp_rej = np.sum(sharpmask)
            logger.debug('%i sources rejected by sharpness',
                         sharp_rej)

        # filter by roundness
        round_rej = 0
        roundmask = False
        if self._round_limit is not None:
            round_limit = sorted(self._round_limit)
            roundmask = np.isnan(sources['round'])
            roundmask |= sources['round'] < round_limit[0]
            roundmask |= sources['round'] > round_limit[1]
            round_rej = np.sum(roundmask)
            logger.debug('%i sources rejected by roundness',
                         round_rej)

        centroidmask = np.isnan(sources['x'])
        centroidmask |= np.isnan(sources['y'])
        logger.debug('%i sources rejected by invalid centroid',
                     np.sum(centroidmask))

        mask = sharpmask | roundmask | centroidmask
        return sources[np.where(~mask)]

    def _group_and_merge(self, sources, image, image_conv, iters=2):
        """Use DBSCAN from sklearn to group and merge close points."""
        # Filter nans
        mask = np.logical_or(np.isnan(sources['x']), np.isnan(sources['y']))
        sources = sources[np.where(~mask)]

        ix = sources['x'].ravel()
        iy = sources['y'].ravel()
        arr = np.transpose([ix, iy])
        # cluster points closer than sigma
        fitted = DBSCAN(eps=self._fwhm/2.355, min_samples=1).fit(arr)
        labels = fitted.labels_
        ngroups = labels.max()+1
        n_sources = np.full(ngroups, dtype=[('x', 'f8'), ('y', 'f8'),
                                            ('flux', 'f8'),
                                            ('round', 'f4'),
                                            ('sharp', 'f8')],
                            fill_value=np.nan)
        n_merged = 0
        for group in range(ngroups):
            rows = sources[np.where(labels == group)]
            n_merged += len(rows)
            if len(rows) > 1:
                xc = int(round(np.mean(rows['x']), 0))
                yc = int(round(np.mean(rows['y']), 0))
                # need to recompute the statistics for the merged source
                sh, rd, f, xcent, ycent = self._compute(image, image_conv,
                                                        xc, yc)
                n_sources[group] = (xcent, ycent, f, rd, sh)
            else:
                n_sources[group] = (rows['x'], rows['y'], rows['flux'],
                                    rows['round'], rows['sharp'])

        logger.debug('merging %i close stars in %i sources',
                     n_merged, ngroups)
        iters -= 1
        if iters:
            n_sources = self._group_and_merge(n_sources, image,
                                              image_conv, iters)
        return n_sources

    def find_stars(self, data, threshold, background, noise):
        """Find stars in a single image using daofind."""
        if len(np.shape(data)) != 2:
            raise ValueError('Data array must be 2 dimensional.')

        n_y, n_x = np.shape(data)
        hmin = threshold*np.median(noise)
        image = np.array(data, dtype=np.float64) - background

        h = self._convolve_image(image, n_x, n_y)

        index = np.where(h >= hmin)
        if len(index[0]) == 0:
            logger.error('No maxima exceed input threshold of %f', hmin)
            return
        logger.debug('Found %i pixels above threshold', len(index[0]))

        index = self._find_peaks(h, index, n_x, n_y)
        t = self._compute_statistics(image, h, index)
        t = self._group_and_merge(t, image, h)
        t = self._filter_sources(t)
        return Table(t)


def daofind(data, threshold, background, noise, fwhm,
            mask=None, **kwargs):
    """Find sources using DAOfind algorithm.

    Parameters
    ----------
    data: array_like
        2D array containing the image to extract the source.
    threshold: `int`
        Minimum signal noise desired.
    background: `int`
        Background estimation.
    noise: `int`
        Root-mean-square at each point.
    fwhm: `int`
        Full width at half maximum: to be used in the convolve filter.
    sharp_limit: array_like or `None` (optional)
        Low and high cutoff for the sharpness statistic.
        `None` will disable the sharpness filtering.
        Default: (0.2, 1.0)
    round_limit : array_like or `None` (optional)
        Low and high cutoff for the roundness statistic.
        `None` will disable the roundness filtering.
        Default: (-1.0,1.0)
    skip_invalid_centroid: `bool`
        If `True`, the code will skip the invalid centroid verification and
        include sources with wrong centroid in the final list.
        Default: `False`
    """
    # TODO: mask is ignored in the original algorith
    # Find a way to use it
    d = DAOFind(fwhm, **kwargs)
    return d.find_stars(data, threshold, background, noise)


def starfind(data, threshold, background, noise, fwhm=None,
             **kwargs):
    """Find stars using daofind AND sepfind.

    Parameters
    ----------
    data: array_like
        2D array containing the image to extract the source.
    threshold: `int`
        Minimum signal noise desired.
    background: `int`
        Background estimation.
    noise: `int`
        Root-mean-square at each point.
    fwhm: `float` (optional)
        Full width at half maximum to be used in the convolve filter.
        No need to be precise and will be recomputed in the function.
        Default: `None`
    mask: `bool` (optional)
        Boolean mask where 1 pixels are masked out in the background
        calculation.
        Default: `None`
    sharp_limit: array_like (optional)
        Low and high cutoff for the sharpness statistic.
        Default: (0.2, 1.0)
    round_limit : array_like (optional)
        Low and high cutoff for the roundness statistic.
        Default: (-1.0,1.0)
    skip_invalid_centroid: `bool`
        If `True`, the code will skip the invalid centroid verification and
        include sources with wrong centroid in the final list.
        Default: `False`
    """
    # First, we identify the sources with sepfind (fwhm independent)
    mask = kwargs.get('mask')
    sources = sepfind(data, threshold, background, noise, mask=mask)

    # We compute the median FWHM and perform a optimum daofind extraction
    box_size = 3*fwhm if fwhm is not None else 15  # 3xFWHM seems to be enough
    min_fwhm = fwhm or 2.0  # hardcoded 3.0 seems to be ok for stars
    fwhm = calc_fwhm(data, sources['x'], sources['y'], box_size=box_size,
                     model='gaussian', min_fwhm=min_fwhm)

    s = daofind(data, threshold, background, noise, fwhm, mask=mask,
                sharp_limit=kwargs.get('sharp_limit', _default_sharp),
                round_limit=kwargs.get('round_limit', _default_round),
                skip_invalid_centroid=kwargs.get('skip_invalid_centroid', 0))
    s.meta['astropop fwhm'] = fwhm
    return s


def sources_mask(shape, x, y, a, b, theta, mask=None, scale=1.0):
    """Create a mask to cover all sources.

    Parameters
    ----------
    shape: int or tuple of ints
        Shape of the new array.
    x,y: array_like
        Center of ellipse(s).
    a, b, theta: array_like (optional)
        Parameters defining the extent of the ellipe(s).
    mask: numpy.ndarray (optional)
        An optional mask.
        Default: `None`
    scale: array_like (optional)
        Scale factor of ellipse(s).
        Default: 1.0
    """
    image = np.zeros(shape, dtype=bool)
    sep.mask_ellipse(image, x, y, a, b, theta, r=scale)
    if mask is not None:
        image |= np.array(mask)
    return image


def _fwhm_loop(model, data, x, y, xc, yc):
    # FIXME: with curve fitting, gaussian model goes crazy
    """
    Parameters
    ----------
    model: `str`
        Choose a Gaussian or Moffat model.
    data: array_like
        2D array containing the image to extract the source.
    x, y: array_like
        x and y indexes ofthe pixels in the image.
    xc, yc: array_like
        x and y initial guess positions of the source.
    """
    if model == 'gaussian':
        model = gaussian_r
        mfwhm = gaussian_fwhm
        p0 = (1.0, np.max(data), np.min(data))
    elif model == 'moffat':
        model = moffat_r
        mfwhm = moffat_fwhm
        p0 = (1.0, 1.5, np.max(data), np.min(data))
    else:
        raise ValueError(f'Model {model} not available.')
    r, f = xy2r(x, y, data, xc, yc)
    args = np.argsort(r)
    try:
        popt, _ = curve_fit(model, r[args], f[args], p0=p0)
        return mfwhm(*popt[:-2])
    except Exception:
        return np.nan


def calc_fwhm(data, x, y, box_size=25, model='gaussian', min_fwhm=3.0):
    """Calculate the median FWHM of the image with Gaussian or Moffat fit.

    Parameters
    ----------
    data: array_like
        2D array containing the image to extract the source.
    x, y: array_like
        x, y centroid position.
    box_size: `int` (optional)
        Size of background boxes in pixels.
        Default: 25
    model: {`gaussian`, `moffat`}
        Choose a Gaussian or Moffat model.
        Default: `gausiann`
    min_fwhm: float
        Minimum value for FWHM.
        Default: 3.0
    """
    indices = np.indices(data.shape)
    rects = [trim_array(data, box_size, (xi, yi), indices=indices)
             for xi, yi in zip(x, y)]
    fwhm = [_fwhm_loop(model, d[0], d[1], d[2], xi, yi)
            for d, xi, yi in zip(rects, x, y)]
    fwhm = np.nanmedian(fwhm)
    if fwhm < min_fwhm or ~np.isfinite(fwhm):
        fwhm = min_fwhm

    return fwhm


def _recenter_loop(fitter, model, data, x, y, xc, yc):
    """ Fit a model in the data to find a new center. Internal only."""
    if model == 'gaussian':
        model = PSFGaussian2D(x_0=xc, y_0=yc)
    elif model == 'moffat':
        model = PSFMoffat2D(x_0=xc, y_0=yc)
    else:
        raise ValueError(f'Model {model} not available.')
    m_fit = fitter(model, x, y, data)
    return m_fit.x_0.value, m_fit.y_0.value


def recenter_sources(data, x, y, box_size=25, model='gaussian'):
    """Recenter teh sources using a PSF model.

    Parameters
    ----------
    data: array_like
        2D array containing the image to extract the source.
    x, y: array_like
        x, y centroid position.
    box_size: `int` (optional)
        Size of background boxes in pixels.
        Default: 25
    model: `str`
        Choose a Gaussian or Moffat model.
        Default: `gausiann`
    """
    # TODO: Failing in tests. Investigate it.
    indices = np.indices(data.shape)
    rects = [trim_array(data, box_size, (xi, yi), indices=indices)
             for xi, yi in zip(x, y)]
    fitter = LevMarLSQFitter()
    coords = [_recenter_loop(fitter, model, d[0], d[1], d[2], xi, yi)
              for d, xi, yi in zip(rects, x, y)]
    coords = np.array(coords)
    nx = coords[:, 0]
    ny = coords[:, 1]
    return nx, ny



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Perform math operations with astronomical images in FrameData contianer."""

import numpy as np
from astropy.units.core import UnitConversionError

from ..framedata import FrameData
from ..math.physical import QFloat, convert_to_qfloat, UnitsError
from ..logger import logger, log_to_list


__all__ = ['imarith']


_arith_funcs = {'+': np.add,
                '-': np.subtract,
                '/': np.true_divide,
                '//': np.floor_divide,
                '*': np.multiply,
                '**': np.power,
                '%': np.remainder}


def _qf_or_framedata(data, alternative=convert_to_qfloat):
    """Check if the data is QFloat or FrameData. Else, convert it."""
    if isinstance(data, (QFloat, FrameData)):
        return data
    return alternative(data)


def _arith_mask(operand1, operand2):
    """Handle the arithmatics of the masks."""
    def _extract(operand):
        if hasattr(operand, 'mask'):
            return operand.mask
        return False

    mask1 = _extract(operand1)
    mask2 = _extract(operand2)

    old_n = np.count_nonzero(mask1)
    nmask = np.logical_or(mask1, mask2)
    new_n = np.count_nonzero(nmask)
    logger.debug('Updating mask in math operation. '
                 'From %i to %i masked elements.', old_n, new_n)
    return nmask


def _arith(operand1, operand2, operation):
    """Perform the math operation itself using QFloats."""
    qf1 = convert_to_qfloat(operand1)
    qf2 = convert_to_qfloat(operand2)

    return _arith_funcs[operation](qf1, qf2)


def _join_headers(operand1, operand2, operation):  # noqa
    """Join the headers to result."""
    # TODO: Think if this is the best behavior
    return operand1.header.copy()


def imarith(operand1, operand2, operation, inplace=False,
            join_masks=False):
    """Perform arithmetic operations using `~astropop.framedata.FrameData`.

    Notes
    -----
    * Keeps the header of the first image.

    * If ``operand1`` is not a `~astropop.framedata.FrameData` instance,
      inplies in ``inplace=False``, and a new `~astropop.framedata.FrameData`
      instance will be created.

    * Supported operations:
        - ``+`` : add. Example: 1+1=2
        - ``-`` : subtract. Example: 2-1=1
        - ``*`` : scalar product. Example: 2*3=6
        - ``/`` : true division. Example: 3/2=1.5
        - ``**`` : power. Example: 3**2=9
        - ``%`` : modulus. Example: 7%2=1
        - ``//`` : floor division. Example: 7//2=3

    Parameters
    ----------
    operand1, operand2: `~astropop.framedata.FrameData` compatible
        Values to perform the operation. `~astropop.math.physical.QFloat`
        `~astropy.units.Quantity`, numerical values and
        `~astropy.nddata.CCDData` are also suported.
    operation: {``+``, ``-``, ``*``, ``/``, ``**``, ``%``, ``//``}
        Math operation.
    inplace: bool, optional
        If True, the operations will be performed inplace in the operand 1.
    join_masks: bool, optional
        Join masks in the end of the operation.

    Returns
    -------
    `~astropop.framedata.FrameData`:
        new `FrameData` instance if not ``inplace``, else the ``operand1``
        `~astropop.framedata.FrameData` instance.
    """
    # TODO: handle WCS
    if operation not in _arith_funcs.keys():
        raise ValueError(f"Operation {operation} not supported.")

    operand1 = _qf_or_framedata(operand1)
    operand2 = _qf_or_framedata(operand2)

    if isinstance(operand1, FrameData) and inplace:
        ccd = operand1
    else:
        ccd = FrameData(None)

    # Add the operation entry to the ccd history.
    lh = log_to_list(logger, ccd.history)
    logger.debug('Operation %s between %s and %s',
                 operation, operand1, operand2)

    # Perform data, mask and uncertainty operations
    try:
        result = _arith(operand1, operand2, operation)
    except UnitConversionError:
        raise UnitsError(f'Units {operand1.unit} and {operand2.unit} are'
                         f' incompatible for {operation} operation.')

    ccd.data = result.nominal
    ccd.unit = result.unit
    ccd.uncertainty = result.uncertainty

    if join_masks:
        ccd.mask = _arith_mask(operand1, operand2)
    else:
        ccd.mask = False

    ccd.meta = _join_headers(operand1, operand2, operation)

    logger.removeHandler(lh)

    return ccd



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Stack and combine astronomical images in FrameData."""

import functools
from tempfile import mkdtemp
import numpy as np
from astropy.stats import mad_std

from ..framedata import FrameData, check_framedata
from ..py_utils import check_iterable, check_number
from ..logger import logger
from ..math.array import all_equal


# TODO: verify why cache_folder is not being used


# bottleneck has faster median
try:
    import bottleneck as bn
    _funcs = {
        'median': bn.nanmedian,
        'mean': bn.nanmean,
        'sum': bn.nansum,
        'std': bn.nanstd
    }
except ImportError:
    _funcs = {
        'median': np.nanmedian,
        'mean': np.nanmean,
        'sum': np.nansum,
        'std': np.nanstd
    }


__all__ = ['imcombine', 'ImCombiner']


_funcs['mad_std'] = functools.partial(mad_std, ignore_nan=True)


def _sigma_clip(data, threshold=3, cen_func='median', dev_func='mad_std',
                axis=None):
    """Create a mask of the sigma clipped pixels.

    This function will not change the array, instead, just output a
    mask for the masked elements.

    Parameters
    ----------
    data: array_like
        Data to be clipped. Can have any dimensionality.
    threshold: int, float or tuple (optional)
        Threshold of sigma clipping. If a number is provided, it is applied
        for both low and high values. If a number is provided, it is
        interpreted as (sigma_low, sigma_high) values.
        Default: 3
    cen_func: callable or {'mean', 'median'} (optional)
        Function to compute the center value used for sigma clipping.
        Default: 'median'
    dev_func: callable or {'std', 'mad_std'} (optional)
        Function to compute the base deviation value used for sigma clipping.
        Default: 'mad_std'
    axis: int
        The axis to perform the clipping and masking.

    Returns
    -------
    mask: `~numpy.ndarray`
        Array with the same shape of `data` containing the mask for elements.
    """
    data = np.array(data)

    if check_number(threshold):
        slow = threshold
        shigh = threshold
    elif check_iterable(threshold):
        slow, shigh = threshold
    else:
        raise TypeError(f'Sigma clipping threshold {threshold} not'
                        ' recognized.')

    if not callable(cen_func):
        cen_func = _funcs[cen_func]
    if not callable(dev_func):
        dev_func = _funcs[dev_func]

    cen = cen_func(data, axis=axis)
    dev = dev_func(data, axis=axis)

    # also mask nans and infs
    mask = ~np.isfinite(data)
    if slow is not None:
        mask |= data < cen-(slow*dev)
    if shigh is not None:
        mask |= data > cen+(shigh*dev)

    logger.debug('Rejected %i pixels by sigmaclip method.',
                 np.sum(mask))

    return mask


def _minmax_clip(data, min_clip=None, max_clip=None):
    """Create a mask of pixels clipped between min_clip and max_clip vals.

    Parameters
    ----------
    data: array_like
        Data array to be cliped.
    min_clip: `float`
        Minimum value accepted in the array. Values lower then this will be
        masked. `None` will disable minimum clip.
        Default: `None`
    max_clip: `float`
        Maximum value accepted in the array. Values greater then this will be
        masked. `None` will disable maximum clip.
        Default: `None`

    Returns
    -------
    mask: `~numpy.ndarray`
        Array with the same shape of `data` containing the mask for elements.
    """
    data = np.array(data)
    # masking nan and infinity
    mask = ~np.isfinite(data)

    if min_clip is not None:
        mask[np.where(data < min_clip)] = True

    if max_clip is not None:
        mask[np.where(data > max_clip)] = True

    logger.debug('Rejected %i pixels by minmax method.',
                 np.sum(mask))

    return mask


class ImCombiner:
    """Process the combining operation of images, like the IRAF imcombine."""

    _sigma_clip = None  # sigmaclip thresholds
    _sigma_cen_func = None  # sigmaclip central function
    _sigma_dev_func = None  # sigmaclip deviation function
    _minmax = None  # minmax clipping parameters
    _max_memory = 1e8  # max memory to be used by the combiner
    _buffer = None  # Temporary buffer to store the image
    _unct_bf = None  # Temporary buffer to store the uncertainties
    _disk_cache = False  # Enable disk caching for _images
    _images = None  # List containing the loaded images
    _methods = {'median', 'mean', 'sum'}
    _dtype = np.float64  # Internal dtype used by the combiner
    _unit = None  # Result unit
    _shape = None  # Full image shape
    _tmpdir = None  # Directory to store temporary files
    _header_merge_keys = None  # Selected header keys for header merging
    _header_strategy = 'no_merge'  # strategy for header merging

    def __init__(self, max_memory=1e9, dtype=np.float64, tmp_dir=None,
                 use_disk_cache=False, **kwargs):
        """Combine images using various algorithms.

        Parameters
        ----------
        max_memory: int (optional)
          Maximum memory to be used during median and mean combining.
          In bytes.
          Default: 1e9 (1GB)
        dtype: `~numpy.dtype` (optional)
          Data type to be used during the operations and the final result.
          Defualt: `~numpy.float64`
        tmp_dir: `str` (optional)
          Directory to store temporary files used in the combining. If None,
          a tmp dir will be created with `~tempfile.mkdtemp`.
          default: `None`
        use_disk_cache: `bool`
            Enable caching images to disk. This slow down the process, but
            avoid memory overflow for large number of images. If the memory
            size needed exceeds max_memory, it is enabled automatically.
            default: false
        merge_header: {'no_merge', 'first', 'only_equal', 'selected_keys'}
            Strategy for merging headers.
            default: 'no_merge'
        merge_header_keys: `list` (optional)
            Keywords for header merging if the strategy is 'selected_keys'.
            default: None
        """
        # workaround to check dtype
        if not isinstance(dtype(0), (float, np.floating)):
            raise ValueError("Only float dtypes are allowed in ImCombiner.")
        self._tmpdir = tmp_dir
        if self._tmpdir is None:
            self._tmpdir = mkdtemp(prefix='astropop')
        self._dtype = dtype
        self._max_memory = max_memory
        # initialize empty image list
        self._images = []
        self._disk_cache = use_disk_cache
        self.set_merge_header(kwargs.pop('merge_header', 'no_merge'),
                              kwargs.pop('merge_header_keys', None))

    def set_sigma_clip(self, sigma_limits=None,
                       center_func='median', dev_func='mad_std'):
        """Enable sigma clipping during the combine.

        Parameters
        ----------
        sigma_limits: `float`, `tuple` or `None` (optional)
          Set the low and high thresholds for sigma clipping. A number is
          applyed to both low and high limits. A tuple will be considered
          (low, high) limits. `None` disable the clipping.
          Default: `None`
        center_func: callable or {'median', 'mean'} (optional)
          Function to compute de central tendency of the data.
          Default: 'median'
        dev_func: callable or {'std', 'mad_std'} (optional)
          Function to compute the deviation sigma for clipping.
          Defautl: 'mad_std'

        Notes
        -----
        - 'median' and 'mad_std' gives a much better sigma clipping than
          'mean' and 'std'.
        """
        if sigma_limits is None:
            # None simply disables sigma clipping
            self._sigma_clip = None
            self._sigma_cen_func = None
            self._sigma_dev_func = None
            return

        if check_iterable(sigma_limits):
            if len(sigma_limits) not in (1, 2):
                raise ValueError('Invalid sigma clipping thresholds'
                                 r' {sigma_limits}')

        if not callable(center_func) and \
           center_func not in ('median', 'mean'):
            raise ValueError(f"Center function {center_func} not accpeted.")

        if not callable(dev_func) and \
           dev_func not in ('std', 'mad_std'):
            raise ValueError(f"Deviation function {dev_func} not accpeted.")

        self._sigma_clip = sigma_limits
        self._sigma_cen_func = center_func
        self._sigma_dev_func = dev_func

    def set_minmax_clip(self, min_value=None, max_value=None):
        """Enable minmax clipping during the combine.

        Parameters
        ----------
        min_value: `float` or `None` (optional)
          Minimum threshold of the clipping. `None` disables minimum masking.
          Default: `None`
        max_value: `float` or `None` (optional)
          Maximum threshold of the clipping. `None` disables maximum masking.
          Default: `None`
        """
        l, h = min_value, max_value
        # disable
        if l is None and h is None:
            self._minmax = None

        for i in (l, h):
            if not check_number(i) and i is not None:
                raise ValueError(f"{i} is not compatible with min_max "
                                 "clipping")

        # check if minimum is lower then maximum
        if l is not None and h is not None:
            l, h = sorted([l, h])

        self._minmax = (l, h)

    def set_merge_header(self, strategy, keys=None):
        """Set the strategy to merge headers during combination.

        Strategies
        ----------
        no_merge:
            No header merging will be done. The resulting combined image
            will contain only new keys based on fits standards.
            Default behavior.
        first:
            All keys from the first header will be used.
        only_equal:
            All keys are compared and just the keys that are equal between
            all hedaers are merged and kept.
        selected_keys:
            Only a list of selected keys by the user will be merged and kept.
            If they are different between headers, the value in the first
            header willbe used.

        Parameters
        ----------
        strategy: {'no_merge', 'first', 'only_equal', 'selected_keys'}
            Header merging strategy.
        keys: list
            List of the keys to be used for `selected_keys` strategy.
        """
        if strategy not in {'no_merge', 'first',
                            'only_equal', 'selected_keys'}:
            raise ValueError(f'{strategy} not known.')

        if strategy == 'selected_keys' and keys is None:
            raise ValueError('No key assigned for `selected_keys` strategy.')

        self._header_strategy = strategy
        self._header_merge_keys = keys

    def _clear(self):
        """Clear buffer and images."""
        self._buffer = None
        # ensure cleaning of tmp files and free memory
        for i in self._images:
            i.disable_memmap()
            i.data = None
        self._images = []
        self._shape = None
        self._unit = None

    def _load_images(self, image_list):
        """Read images to FrameData and enable memmap."""
        # clear the buffers before load images.
        self._clear()

        if len(image_list) == 0:
            raise ValueError('Image list is empty.')

        for indx, i in enumerate(image_list):
            # before combine, copy everything to FrameData
            ic = check_framedata(i, copy=True)
            ic = ic.astype(self._dtype)

            # for optimization, only enable memmap when needed.
            if self._disk_cache:
                ic.enable_memmap(filename=f'image_{indx}',
                                 cache_folder=self._tmpdir)

            self._images.append(ic)

    def _check_consistency(self):
        """Check the consistency between loaded images."""
        if len(self._images) == 0:
            raise ValueError('Combiner have no images.')
        base_shape = None
        base_unit = None
        for i, v in enumerate(self._images):
            # supose self._images only have FrameData beacuse it's protected
            if i == 0:
                base_shape = v.shape
                base_unit = v.unit
            elif v.shape != base_shape:
                raise ValueError(f"Image {i} has a shape incompatible with "
                                 "the others")
            elif v.unit != base_unit:
                raise ValueError(f"Image {i} has a unit incompatible with "
                                 "the others")
        self._shape = base_shape
        self._unit = base_unit

    def _chunk_yielder(self, method):
        """Split the data in chuncks according to the method."""
        # sum needs uncertainties
        unct = None
        if method == 'sum':
            if not np.any([i.uncertainty.empty for i in self._images]):
                unct = True
            else:
                logger.info('One or more frames have empty uncertainty. '
                            'Some features are disabled.')

        shape = self._images[0].shape
        tot_size = self._images[0].data.nbytes
        tot_size += self._images[0].mask.nbytes
        tot_size *= len(self._images)
        # uncertainty is ignored

        # adjust memory usage for numpy and bottleneck, based on ccdproc
        if method == 'median':
            tot_size *= 4.5
        else:
            tot_size *= 3

        n_chunks = np.ceil(tot_size/self._max_memory)

        # compute x and y steps
        xstep = max(1, int(shape[0]/n_chunks))
        if shape[0] >= n_chunks:
            ystep = shape[1]
        else:
            ystep = max(1, int(np.ceil(shape[1]/(n_chunks/shape[0]))))

        n_chunks = np.ceil(shape[0]/xstep)*np.ceil(shape[1]/ystep)
        if n_chunks > 1:
            logger.debug('Splitting the images into %i chunks.', n_chunks)

        for x in range(0, shape[0], xstep):
            for y in range(0, shape[1], ystep):
                slc_x = slice(x, min(x+xstep, shape[0]))
                slc_y = slice(y, min(y+ystep, shape[1]))
                shp = slc_x.stop-slc_x.start, slc_y.stop-slc_y.start
                slc = (slc_x, slc_y)
                buffer = np.full((len(self._images), shp[0], shp[1]),
                                 fill_value=np.nan, dtype=self._dtype)
                if unct:
                    unct_buffer = np.full((len(self._images), shp[0], shp[1]),
                                          fill_value=np.nan, dtype=self._dtype)
                else:
                    unct_buffer = None

                for i in range(len(self._images)):
                    buffer[i] = self._images[i].data[slc]
                    buffer[i][self._images[i].mask[slc]] = np.nan
                    if unct:
                        unct_buffer[i] = self._images[i].uncertainty[slc]
                        unct_buffer[i][self._images[i].mask[slc]] = np.nan

                yield buffer, unct_buffer, (slc_x, slc_y)

    def _apply_rejection(self):
        mask = np.zeros(self._buffer[0].shape)
        if self._sigma_clip is not None:
            mask = _sigma_clip(self._buffer,
                               threshold=self._sigma_clip,
                               cen_func=self._sigma_cen_func,
                               dev_func=self._sigma_dev_func,
                               axis=0)

        if self._minmax is not None:
            _min, _max = self._minmax
            mask = np.logical_or(_minmax_clip(self._buffer, _min, _max),
                                 mask)

        mask = np.logical_or(np.isnan(self._buffer), mask)
        self._buffer[mask] = np.nan

    def _combine(self, method, **kwargs):
        """Process the combine and compute the uncertainty."""
        # number of masked pixels for each position
        n_masked = np.sum([np.isnan(i) for i in self._buffer], axis=0)
        # number of images
        n = float(len(self._buffer))
        # number of not masked pixels for each position
        n_no_mask = n - n_masked

        if method == 'sum':
            data = _funcs['sum'](self._buffer, axis=0)
            if self._unct_bf is None:
                logger.info('Data with no uncertainties. Using the std dev'
                            ' approximation to compute the sum uncertainty.')
                # we consider, here, that the deviation in each pixel (x, y) is
                # the error of each image in that position. So
                # unct = stddev*sqrt(n)
                unct = _funcs['std'](self._buffer, axis=0)*np.sqrt(n_no_mask)
            else:
                # direct propagate the errors in the sum
                # unct = sqrt(sigma1^2 + sigma2^2 + ...)
                unct = _funcs['sum'](np.square(self._unct_bf), axis=0)
                unct = np.sqrt(unct)

            if kwargs.get('sum_normalize', True):
                norm = n/n_no_mask
                data *= norm
                unct *= norm

        elif method in ('median', 'mean'):
            data = _funcs[method](self._buffer, axis=0)
            # uncertainty = sigma/sqrt(n)
            unct = _funcs['std'](self._buffer, axis=0)
            unct /= np.sqrt(n_no_mask)

        return data, unct

    def _merge_header(self):
        """Merge headers."""
        meta = {}
        if self._header_strategy == 'no_merge':
            return meta

        logger.debug('Merging headers with %s strategy.',
                     self._header_strategy)

        if self._header_strategy == 'first':
            return self._images[0].header

        summary = {h: [] for h in self._images[0].header.keys()}
        for i in self._images:
            hdr = i.header
            for key in hdr.keys():
                if key not in summary.keys():
                    summary[key] = []
                if hdr[key] not in summary[key]:
                    summary[key].append(hdr[key])

        if self._header_strategy == 'selected_keys':
            keys = self._header_merge_keys
        else:
            keys = summary.keys()

        for k in keys:
            if all_equal(np.array(summary[k])):
                meta[k] = summary[k][0]
            elif self._header_strategy == 'selected_keys':
                logger.debug('Keyword %s is different across headers. '
                             'Unsing first one.', k)
                meta[k] = summary[k][0]

        return meta

    def combine(self, image_list, method, **kwargs):
        """Perform the image combining.

        Parameters
        ----------
        image_list: `list` or `tuple`
            List containing the images to be combined. The values in the list
            must be all of the same type and `~astropop.framedata.FrameData`
            supported.
        method: {'mean', 'median', 'sum'}
            Combining method.
        **kwargs:
            sum_normalize: bool (optional)
                If True, the imaged will be multiplied, pixel by pixel, by the
                number of images divided by the number of non-masked pixels.
                This will avoid discrepancies by different numbers of masked
                pixels across the image. If False, the raw sum of images will
                be returned.
                Default: True

        Returns
        -------
        combined: `~astropop.framedata.FrameData`
          The combined image.

        Notes
        -----
        - For now, it don't consider WCS, so it perform plain between the
          images, whitout registering.
        - Clipping parameters are set using class functions.
        - If the images exceed the maximum memory allowed, they are splited
          to perform the median and mean combine.
        - Masked elements are skiped. Result pixels will be masked if all the
          source pixels combined in it are also masked.
        """
        if method not in self._methods:
            raise ValueError(f'{method} is not a valid combining method.')

        # first of all, load the images to FrameData and check the consistency
        self._load_images(image_list)
        self._check_consistency()

        logger.info('Combining %i images with %s method.',
                    len(self._images), method)

        # temp combined data, mask and uncertainty
        data = np.zeros(self._shape, dtype=self._dtype)
        data.fill(np.nan)
        mask = np.zeros(self._shape, dtype=bool)
        unct = np.zeros(self._shape, dtype=self._dtype)

        for self._buffer, self._unct_bf, slc in self._chunk_yielder(method):
            # perform the masking: first with minmax, after sigma_clip
            # the clippings interfere in each other.
            self._apply_rejection()

            # combine the images and compute the uncertainty
            data[slc], unct[slc] = self._combine(method, **kwargs)

            # combine masks
            mask[slc] = np.all([np.isnan(i) for i in self._buffer], axis=0)
            # TODO: flag pixels with NaN as pixels with rejection

        n = len(self._images)
        combined = FrameData(data, unit=self._unit, mask=mask,
                             uncertainty=unct)
        combined.meta = self._merge_header()
        combined.meta['HIERARCH astropop imcombine nimages'] = n
        combined.meta['HIERARCH astropop imcombine method'] = method

        # after, clear all buffers
        self._clear()
        return combined


def imcombine(frames, method='median', memory_limit=1e9, **kwargs):
    """Combine a list of images or frames in a single one.

    Parameters
    ----------
    frames: list
        List of the frames to be combined. Can be a list of `FrameData`,
        a list of file names, a list of `~astropy.fits.ImageHDU` or
        a list of `~numpy.ndarray`. All members must have the same dimensions.
        For `FrameData`, all units must be the compatible.
    method: {'sum', 'median', 'mean'}
        Combining method. Each one has a proper math and a proper error
        computation.
    memory_limit: int (optional)
        The maximum memory limit (in bytes) to be used in the combining.
        If the data exceeds the maximum memory limit, it will be slipted in
        chunks for the rejection and combining processes.
    **kwargs:
        sigma_clip: float or tuple (optional)
            Threshold of sigma clipping rejection. If `None`, it disables the
            sigma clipping. If a number is provided, it is applied for both low
            and high values. If a tuple is provided, it is interpreted as
            (sigma_low, sigma_high) values.
            Default: `None`
        sigma_cen_func: callable or {'median', 'mean'} (optional)
            Function to compute the central value of sigma clipping rejection.
            If a name is provided, it must follow the convention in Notes.
            If a callable is provided, it will be applied directly on the data
            and must accept 'axis' argument.
            Default: 'median'
        sigma_dev_func: callable or {'std', 'mad_std'} (optional)
            Function to compute the std deviation of sigma clipping rejection.
            If a name is provided, it must follow the convention in Notes.
            If a callable is provided, it will be applied directly on the data
            and must accept 'axis' argument.
            Default: 'std'
        minmax_clip: tuple (optional)
            Minimum and maximum limits for minmax clipping. The values are
            interpreted as (min, max) limits. All values lower then the minimum
            limit and greater then the maximum limit will be masked. If `None`,
            the minmax clipping will be disabled.
            Default: `None`
        merge_header: {'first', 'selected_keys', 'no_merge'} (optional)
            Strategy to merge the headers of the images. If 'first', the first
            image header will be used. If 'selected_keys', the headers will be
            merged using the keys provided in the `header_merge_keys`.
            If 'no_merge', the headers will not be merged.
            Default: 'no_merge'
        header_merge_keys: list (optional)
            List of the keywords to be used to merge the headers.
            Defaut: `None`

    Returns
    -------
    result: `FrameData`
        A `FrameData` containing the combined image and its uncertainty.

    Notes
    -----
    - It is not recomended using clipping with 'sum' method, since it will
      change the number of elements to be summed in each column.
    - To disable a low or high clipping, use tuple with None. For example,
      using `sigma_clip=(None, 2)`, the lower clipping will be disabled.
    - The center function names are:
      - 'median': `~numpy.nanmedian`
      - 'mean': `~numpy.nanmean`
    - The standard deviation funcion names are:
      - 'std': `~numpy.nanstd`
      - 'mad_std': `~astropy.stats.funcs.mad_std`
    """
    # Sanitize kwargs and create combiner
    kargs = {'dtype': kwargs.pop('dtype', np.float64),
             'use_disk_cache': kwargs.pop('use_disk_cache', False),
             'tmp_dir': kwargs.pop('tmp_dir', None)}
    combiner = ImCombiner(max_memory=memory_limit, **kargs)

    sc = kwargs.pop('sigma_clip', None)
    if sc is not None:
        sigma_cen_func = kwargs.pop('sigma_cen_func', 'median')
        sigma_dev_func = kwargs.pop('sigma_dev_func', 'std')
        combiner.set_sigma_clip(sc, sigma_cen_func, sigma_dev_func)

    mm = kwargs.pop('minmax_clip', None)
    combiner.set_minmax_clip(mm)

    merge_header = kwargs.pop('merge_header', 'no_merge')
    header_keys = kwargs.pop('merge_header_keys', None)
    combiner.set_merge_header(merge_header, header_keys)

    # Perform the combinations
    return combiner.combine(frames, method, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
from astropop._db import SQLColumn, SQLRow, SQLTable, _ID_KEY, \
                         SQLDatabase, _sanitize_colnames, \
                         SQLColumnMap
import numpy as np
from astropy.table import Table
from astropop.testing import *
import sqlite3


def test_sanitize_string():
    for i in ['test-2', 'test!2', 'test@2', 'test#2', 'test$2',
              'test&2', 'test*2', 'test(2)', 'test)2', 'test[2]', 'test]2',
              'test{2}', 'test}2', 'test|2', 'test\\2', 'test^2', 'test~2'
              'test"2', 'test\'2', 'test`2', 'test<2', 'test>2', 'test=2',
              'test,2', 'test;2', 'test:2', 'test?2', 'test/2']:
        with pytest.raises(ValueError):
            _sanitize_colnames(i)

    for i in ['test', 'test_1', 'test_1_2', 'test_1_2', 'Test', 'Test_1']:
        assert_equal(_sanitize_colnames(i), i.lower())


class Test_SQLColumnMap:
    def cmap(self):
        db = SQLDatabase()
        db.add_table('key_columns', data={'keywords': ['key1', 'key 2',
                                                       'key-3', 'key_4'],
                                          'columns': ['col1', 'col2', 'col3',
                                                      'col4']})
        cmap = SQLColumnMap(db, 'key_columns', 'keywords', 'columns')
        return cmap

    def test_columnmap_get_column_name(self):
        cmap = self.cmap()
        assert_equal(cmap.get_column_name('key1'), 'col1')
        assert_equal(cmap.get_column_name('key 2'), 'col2')
        assert_equal(cmap.get_column_name('key-3'), 'col3')
        assert_equal(cmap.get_column_name('key_4'), 'col4')

    def test_columnmap_get_column_name_list(self):
        cmap = self.cmap()
        assert_equal(cmap.get_column_name(['key1', 'key 2']), ['col1', 'col2'])

    def test_columnmap_get_column_name_not_found(self):
        cmap = self.cmap()
        with pytest.raises(KeyError):
            cmap.get_column_name('key5')

    def test_columnmap_get_keyword(self):
        cmap = self.cmap()
        assert_equal(cmap.get_keyword('col1'), 'key1')
        assert_equal(cmap.get_keyword('col2'), 'key 2')
        assert_equal(cmap.get_keyword('col3'), 'key-3')
        assert_equal(cmap.get_keyword('col4'), 'key_4')

    def test_columnmap_get_keyword_list(self):
        cmap = self.cmap()
        assert_equal(cmap.get_keyword(['col1', 'col2']), ['key1', 'key 2'])

    def test_columnmap_get_keyword_not_found(self):
        db = SQLDatabase()
        db.add_table('key_columns', data={'keywords': ['key1', 'key 2',
                                                       'key-3', 'key_4'],
                                          'columns': ['col1', 'col2', 'col3',
                                                      'col4']})
        cmap = SQLColumnMap(db, 'key_columns', 'keywords', 'columns')
        with pytest.raises(KeyError):
            cmap.get_keyword('col5')

    def test_columnmap_map_row(self):
        cmap = self.cmap()
        row = {'key1': 1, 'key 2': 2, 'key-3': 3}
        assert_equal(cmap.map_row(row), {'col1': 1, 'col2': 2, 'col3': 3})

    def test_columnmap_map_row_add_column(self):
        cmap = self.cmap()
        row = {'key1': 1, 'key 2': 2, 'key 5': 3}
        assert_equal(cmap.map_row(row, add_columns=True),
                     {'col1': 1, 'col2': 2, 'col_5': 3})

    def test_columnmap_map_row_list(self):
        cmap = self.cmap()
        row = [1, 2, 3, 4]
        assert_equal(cmap.map_row(row), [1, 2, 3, 4])

    def test_columnmap_add_column(self):
        cmap = self.cmap()
        cmap.add_column('key 55')
        assert_equal(cmap.get_column_name('key 55'), 'col_5')

    def test_columnmap_add_column_existing(self):
        cmap = self.cmap()
        with pytest.raises(ValueError):
            cmap.add_column('key1')

    def test_columnmap_parse_where(self):
        cmap = self.cmap()
        with pytest.raises(TypeError):
            cmap.parse_where('key1 = 1')
        with pytest.raises(TypeError):
            cmap.parse_where(['key1 = 1', 'key2 = 2'])
        cmap.parse_where({'key1': 1, 'key 2': 2})


class Test_SQLDatabase_Creation_Modify:
    def test_sql_db_creation(self):
        db = SQLDatabase(':memory:')
        assert_equal(db.table_names, [])
        assert_equal(len(db), 0)

        db.add_table('test')
        assert_equal(db.table_names, ['test'])
        assert_equal(len(db), 1)

    def test_sql_add_column_name_and_data(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db.get_column('test', 'a').values, np.arange(10, 20))
        assert_equal(db.get_column('test', 'b').values, np.arange(20, 30))
        assert_equal(db.column_names('test'), ['a', 'b'])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_column_only_name(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a')
        db.add_column('test', 'b')

        assert_equal(db.get_column('test', 'a').values, [])
        assert_equal(db.get_column('test', 'b').values, [])
        assert_equal(db.column_names('test'), ['a', 'b'])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_table_from_data_table(self):
        db = SQLDatabase(':memory:')
        d = Table(names=['a', 'b'], data=[np.arange(10, 20), np.arange(20, 30)])
        db.add_table('test', data=d)

        assert_equal(db.get_column('test', 'a').values, np.arange(10, 20))
        assert_equal(db.get_column('test', 'b').values, np.arange(20, 30))
        assert_equal(db.column_names('test'), ['a', 'b'])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_table_from_data_ndarray(self):
        dtype = [('a', 'i4'), ('b', 'f8')]
        data = np.array([(1, 2.0), (3, 4.0), (5, 6.0), (7, 8.0)], dtype=dtype)
        db = SQLDatabase(':memory:')
        db.add_table('test', data=data)

        assert_equal(db.get_column('test', 'a').values, [1, 3, 5, 7])
        assert_equal(db.get_column('test', 'b').values, [2.0, 4.0, 6.0, 8.0])
        assert_equal(db.column_names('test'), ['a', 'b'])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_table_from_data_dict(self):
        d = {'a': np.arange(10, 20), 'b': np.arange(20, 30)}
        db = SQLDatabase(':memory:')
        db.add_table('test', data=d)

        assert_equal(db.get_column('test', 'a').values, np.arange(10, 20))
        assert_equal(db.get_column('test', 'b').values, np.arange(20, 30))
        assert_equal(db.column_names('test'), ['a', 'b'])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_table_from_data_ndarray_untyped(self):
        # Untyped ndarray should fail in get column names
        data = np.array([(1, 2.0), (3, 4.0), (5, 6.0), (7, 8.0)])
        db = SQLDatabase(':memory:')
        with pytest.raises(ValueError):
            db.add_table('test', data=data)

    @pytest.mark.parametrize('data, error', [([1, 2, 3], ValueError),
                                             (1, TypeError),
                                             (1.0, TypeError),
                                             ('test', TypeError)])
    def test_sql_add_table_from_data_invalid(self, data, error):
        db = SQLDatabase(':memory:')
        with assert_raises(error):
            db.add_table('test', data=data)

    def test_sql_add_table_columns(self):
        db = SQLDatabase(':memory:')
        db.add_table('test', columns=['a', 'b'])

        assert_equal(db.get_column('test', 'a').values, [])
        assert_equal(db.get_column('test', 'b').values, [])
        assert_equal(db.column_names('test'), ['a', 'b'])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_table_columns_data(self):
        db = SQLDatabase(':memory:')
        with pytest.raises(ValueError):
            db.add_table('test', columns=['a', 'b'], data=[1, 2, 3])

    def test_sql_add_row(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a')
        db.add_column('test', 'b')
        db.add_rows('test', dict(a=1, b=2))
        db.add_rows('test', dict(a=[3, 5], b=[4, 6]))

        assert_equal(db.get_column('test', 'a').values, [1, 3, 5])
        assert_equal(db.get_column('test', 'b').values, [2, 4, 6])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_row_types(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        for k in ['a', 'b', 'c', 'd', 'e']:
            db.add_column('test', k)

        db.add_rows('test', dict(a=1, b='a', c=True, d=b'a', e=3.14))
        db.add_rows('test', dict(a=2, b='b', c=False, d=b'b', e=2.71))

        assert_equal(db.get_column('test', 'a').values, [1, 2])
        assert_equal(db.get_column('test', 'b').values, ['a', 'b'])
        assert_equal(db.get_column('test', 'c').values, [1, 0])
        assert_equal(db.get_column('test', 'd').values, [b'a', b'b'])
        assert_almost_equal(db.get_column('test', 'e').values, [3.14, 2.71])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])

    def test_sql_add_row_invalid(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a')
        db.add_column('test', 'b')
        with assert_raises(ValueError):
            db.add_rows('test', [1, 2, 3])

    def test_sql_add_row_add_columns(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a')
        db.add_column('test', 'b')
        db.add_rows('test', dict(a=1, b=2))
        db.add_rows('test', dict(a=3, c=4), add_columns=False)
        db.add_rows('test', dict(a=5, d=6), add_columns=True)

        assert_equal(db.get_column('test', 'a').values, [1, 3, 5])
        assert_equal(db.get_column('test', 'b').values, [2, None, None])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])
        assert_equal(db.column_names('test'), ['a', 'b', 'd'])

    def test_sql_add_row_superpass_64_limit(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_rows('test', {f'col{i}': np.arange(10) for i in range(128)},
                     add_columns=True)
        assert_equal(db.column_names('test'), [f'col{i}' for i in range(128)])

        for i, v in enumerate(db.select('test')):
            assert_equal(v, [i]*128)

    def test_sqltable_add_column(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db['test'].add_column('a')
        db['test'].add_column('b')
        db['test'].add_column('c', data=[1, 2, 3])

        assert_equal(db.get_column('test', 'a').values, [None, None, None])
        assert_equal(db.get_column('test', 'b').values, [None, None, None])
        assert_equal(db.get_column('test', 'c').values, [1, 2, 3])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])
        assert_equal(db.column_names('test'), ['a', 'b', 'c'])
        assert_equal(len(db['test']), 3)

    def test_sqltable_add_row_add_columns(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a')
        db.add_column('test', 'b')
        db['test'].add_rows(dict(a=1, b=2))
        db['test'].add_rows(dict(a=3, c=4), add_columns=False)
        db['test'].add_rows(dict(a=5, d=6), add_columns=True)

        assert_equal(db.get_column('test', 'a').values, [1, 3, 5])
        assert_equal(db.get_column('test', 'b').values, [2, None, None])
        assert_equal(len(db), 1)
        assert_equal(db.table_names, ['test'])
        assert_equal(db.column_names('test'), ['a', 'b', 'd'])

    def test_sql_set_column(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', [1, 3, 5])
        db.add_column('test', 'b', [2, 4, 6])

        db.set_column('test', 'a', [10, 20, 30])
        db.set_column('test', 'b', [20, 40, 60])

        assert_equal(db.get_column('test', 'a').values, [10, 20, 30])
        assert_equal(db.get_column('test', 'b').values, [20, 40, 60])

        with pytest.raises(KeyError):
            db.set_column('test', 'c', [10, 20, 30])
        with pytest.raises(ValueError):
            db.set_column('test', 'a', [10, 20, 30, 40])

    def test_sql_set_row(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', [1, 3, 5])
        db.add_column('test', 'b', [2, 4, 6])

        db.set_row('test', 0, dict(a=10, b=20))
        db.set_row('test', 1, [20, 40])
        db.set_row('test', 2, np.array([30, 60]))

        assert_equal(db.get_column('test', 'a').values, [10, 20, 30])
        assert_equal(db.get_column('test', 'b').values, [20, 40, 60])

        with pytest.raises(IndexError):
            db.set_row('test', 3, dict(a=10, b=20))
        with pytest.raises(IndexError):
            db.set_row('test', -4, dict(a=10, b=20))

    def test_sql_set_item(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', [1, 3, 5])
        db.add_column('test', 'b', [2, 4, 6])

        db.set_item('test', 'a', 0, 10)
        db.set_item('test', 'b', 1, 'a')
        assert_equal(db.get_column('test', 'a').values, [10, 3, 5])
        assert_equal(db.get_column('test', 'b').values, [2, 'a', 6])

    def test_sql_setitem_tuple_only(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        with pytest.raises(KeyError):
            db[1] = 0
        with pytest.raises(KeyError):
            db['notable'] = 0
        with pytest.raises(KeyError):
            db[['test', 0]] = 0
        with pytest.raises(KeyError):
            db[1, 0] = 0

    def test_sql_setitem(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        db['test', 'a'] = np.arange(50, 60)
        db['test', 0] = {'a': 1, 'b': 2}
        db['test', 'b', 5] = -999

        expect = np.transpose([np.arange(50, 60), np.arange(20, 30)])
        expect[0] = [1, 2]
        expect[5, 1] = -999

        assert_equal(db['test'].values, expect)

    def test_sql_droptable(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', [1, 3, 5])
        db.add_column('test', 'b', [2, 4, 6])

        db.drop_table('test')
        assert_equal(db.table_names, [])
        with pytest.raises(KeyError):
            db['test']

    def test_sql_copy(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', [1, 3, 5])
        db.add_column('test', 'b', [2, 4, 6])

        db2 = db.copy()
        assert_equal(db2.table_names, ['test'])
        assert_equal(db2.column_names('test'), ['a', 'b'])
        assert_equal(db2.get_column('test', 'a').values, [1, 3, 5])
        assert_equal(db2.get_column('test', 'b').values, [2, 4, 6])

    def test_sql_copy_indexes(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', np.arange(1, 101, 2))
        db.add_column('test', 'b', np.arange(2, 102, 2))

        db2 = db.copy(indexes={'test': [30, 24, 32, 11]})
        assert_equal(db2.table_names, ['test'])
        assert_equal(db2.column_names('test'), ['a', 'b'])
        assert_equal(db2.get_column('test', 'a').values, [23, 49, 61, 65])
        assert_equal(db2.get_column('test', 'b').values, [24, 50, 62, 66])


class Test_SQLDatabase_Access:
    def test_sql_get_table(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db.get_table('test').values, list(zip(np.arange(10, 20),
                                                           np.arange(20, 30))))
        assert_is_instance(db.get_table('test'), SQLTable)

        with pytest.raises(KeyError):
            db.get_table('not_a_table')

    def test_sql_get_table_empty(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')

        assert_equal(len(db.get_table('test')), 0)
        assert_is_instance(db.get_table('test'), SQLTable)

    def test_sql_get_column(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db.get_column('test', 'a').values, np.arange(10, 20))
        assert_equal(db.get_column('test', 'b').values, np.arange(20, 30))
        assert_is_instance(db.get_column('test', 'a'), SQLColumn)
        assert_is_instance(db.get_column('test', 'b'), SQLColumn)

        # same access from table
        assert_equal(db.get_table('test').get_column('a').values, np.arange(10, 20))
        assert_equal(db.get_table('test').get_column('b').values, np.arange(20, 30))
        assert_is_instance(db.get_table('test').get_column('a'), SQLColumn)
        assert_is_instance(db.get_table('test').get_column('b'), SQLColumn)

        with pytest.raises(KeyError):
            db.get_column('test', 'c')
        with pytest.raises(KeyError):
            db.get_table('test').get_column('c')

    def test_sql_get_row(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db.get_row('test', 4).values, (14, 24))
        assert_is_instance(db.get_row('test', 4), SQLRow)

        assert_equal(db.get_row('test', -1).values, (19, 29))
        assert_is_instance(db.get_row('test', -1), SQLRow)

        # same access from table
        assert_equal(db.get_table('test').get_row(4).values, [14, 24])
        assert_is_instance(db.get_table('test').get_row(4), SQLRow)

        with pytest.raises(IndexError):
            db.get_row('test', 11)
        with pytest.raises(IndexError):
            db.get_row('test', -11)
        with pytest.raises(IndexError):
            db.get_table('test').get_row(11)
        with pytest.raises(IndexError):
            db.get_table('test').get_row(-11)

    def test_sql_getitem(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db['test']['a'].values, np.arange(10, 20))
        assert_equal(db['test']['b'].values, np.arange(20, 30))
        assert_is_instance(db['test']['a'], SQLColumn)
        assert_is_instance(db['test']['b'], SQLColumn)

        assert_equal(db['test'][4].values, (14, 24))
        assert_is_instance(db['test'][4], SQLRow)
        assert_equal(db['test'][-1].values, (19, 29))
        assert_is_instance(db['test'][-1], SQLRow)

        with pytest.raises(KeyError):
            db['test']['c']
        with pytest.raises(KeyError):
            db['not_a_table']['a']

        with pytest.raises(IndexError):
            db['test'][11]
        with pytest.raises(IndexError):
            db['test'][-11]

    def test_sql_getitem_tuple(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db['test', 'a'].values, np.arange(10, 20))
        assert_equal(db['test', 'b'].values, np.arange(20, 30))
        assert_is_instance(db['test', 'a'], SQLColumn)
        assert_is_instance(db['test', 'b'], SQLColumn)

        assert_equal(db['test', 4].values, (14, 24))
        assert_is_instance(db['test', 4], SQLRow)
        assert_equal(db['test', -1].values, (19, 29))
        assert_is_instance(db['test', -1], SQLRow)

        assert_equal(db['test', 'a', 4], 14)
        assert_equal(db['test', 'b', 4], 24)
        assert_equal(db['test', 'a', -1], 19)
        assert_equal(db['test', 'b', -1], 29)
        assert_equal(db['test', 4, 'a'], 14)
        assert_equal(db['test', 4, 'b'], 24)
        assert_equal(db['test', -1, 'a'], 19)
        assert_equal(db['test', -1, 'b'], 29)

    def test_sql_getitem_table_force(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        with pytest.raises(ValueError):
            db[1]
        with pytest.raises(ValueError):
            db[1, 2]
        with pytest.raises(ValueError):
            db[1, 2, 'test']
        with pytest.raises(ValueError):
            db[[1, 2], 'test']


class Test_SQLDatabase_PropsComms:
    def test_sql_select_where(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        a = db.select('test', columns='a', where={'a': 15})
        assert_equal(a, 15)

        a = db.select('test', columns=['a', 'b'], where={'b': 22})
        assert_equal(a, [(12, 22)])

        a = db.select('test', columns=['a', 'b'], where=None)
        assert_equal(a, list(zip(np.arange(10, 20), np.arange(20, 30))))

        a = db.select('test', columns=['a', 'b'], where=['a > 12', 'b < 26'])
        assert_equal(a, [(13, 23), (14, 24), (15, 25)])

    def test_sql_select_limit_offset(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        a = db.select('test', columns='a', limit=1)
        assert_equal(a, 10)

        a = db.select('test', columns='a', limit=3, offset=2)
        assert_equal(a, [[12], [13], [14]])

    def test_sql_select_invalid(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        with pytest.raises(sqlite3.OperationalError,
                           match='no such column: c'):
            db.select('test', columns=['c'])

        with pytest.raises(ValueError,
                           match='offset cannot be used without limit.'):
            db.select('test', columns='a', offset=1)

        with pytest.raises(TypeError, match='where must be'):
            db.select('test', columns='a', where=1)

        with pytest.raises(TypeError, match='if where is a list'):
            db.select('test', columns='a', where=[1, 2, 3])

        with pytest.raises(TypeError):
            db.select('test', limit=3.14)

        with pytest.raises(TypeError):
            db.select('test', order=5)

    def test_sql_select_order(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30)[::-1])

        a = db.select('test', order='b')
        assert_equal(a, list(zip(np.arange(10, 20),
                                 np.arange(20, 30)[::-1]))[::-1])

        a = db.select('test', order='b', limit=2)
        assert_equal(a, [(19, 20), (18, 21)])

        a = db.select('test', order='b', limit=2, offset=2)
        assert_equal(a, [(17, 22), (16, 23)])

        a = db.select('test', order='b', where='a < 15')
        assert_equal(a, [(14, 25), (13, 26), (12, 27), (11, 28), (10, 29)])


        a = db.select('test', order='b', where='a < 15', limit=3)
        assert_equal(a, [(14, 25), (13, 26), (12, 27)])

        a = db.select('test', order='b', where='a < 15', limit=3, offset=2)
        assert_equal(a, [(12, 27), (11, 28), (10, 29)])

    def test_sql_count(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db.count('test'), 10)
        assert_equal(db.count('test', where={'a': 15}), 1)
        assert_equal(db.count('test', where={'a': 15, 'b': 22}), 0)
        assert_equal(db.count('test', where='a > 15'), 4)
        assert_equal(db.count('test', where=['a > 15', 'b < 27']), 1)

    def test_sql_prop_db(self, tmp_path):
        db = SQLDatabase(':memory:')
        assert_equal(db.db, ':memory:')

        db = SQLDatabase(str(tmp_path / 'test.db'))
        assert_equal(db.db, str(tmp_path / 'test.db'))

    def test_sql_prop_table_names(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_table('test2')
        assert_equal(db.table_names, ['test', 'test2'])

    def test_sql_prop_column_names(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))
        assert_equal(db.column_names('test'), ['a', 'b'])

    def test_sql_repr(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))
        db.add_table('test2')
        db.add_column('test2', 'a', data=np.arange(10, 20))
        db.add_column('test2', 'b', data=np.arange(20, 30))

        expect = f"SQLDatabase ':memory:' at {hex(id(db))}:\n"
        expect += "\ttest: 2 columns 10 rows\n"
        expect += "\ttest2: 2 columns 10 rows"
        assert_equal(repr(db), expect)

        db = SQLDatabase(':memory:')
        expect = f"SQLDatabase ':memory:' at {hex(id(db))}:\n"
        expect += "\tEmpty database."
        assert_equal(repr(db), expect)

    def test_sql_len(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_table('test2')
        db.add_table('test3')

        assert_equal(len(db), 3)

    def test_sql_index_of(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        assert_equal(db.index_of('test', {'a': 15}), 5)
        assert_equal(db.index_of('test', 'b >= 27'), [7, 8, 9])
        assert_equal(db.index_of('test', {'a': 1, 'b': 2}), [])


class Test_SQLRow:
    @property
    def db(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))
        return db

    def test_row_basic_properties(self):
        db = self.db
        row = db['test'][0]
        assert_is_instance(row, SQLRow)
        assert_equal(row.table, 'test')
        assert_equal(row.index, 0)
        assert_equal(row.column_names, ['a', 'b'])
        assert_equal(row.keys, ['a', 'b'])
        assert_equal(row.values, [10, 20])
        assert_equal(row.as_dict(), {'a': 10, 'b': 20})

    def test_row_iter(self):
        db = self.db
        row = db['test'][0]
        assert_is_instance(row, SQLRow)

        v = 10
        for i in row:
            assert_equal(i, v)
            v += 10

    def test_row_getitem(self):
        db = self.db
        row = db['test'][0]
        assert_is_instance(row, SQLRow)

        assert_equal(row['a'], 10)
        assert_equal(row['b'], 20)

        with pytest.raises(KeyError):
            row['c']

        assert_equal(row[0], 10)
        assert_equal(row[1], 20)
        assert_equal(row[-1], 20)
        assert_equal(row[-2], 10)

        with pytest.raises(IndexError):
            row[2]
        with pytest.raises(IndexError):
            row[-3]

    def test_row_setitem(self):
        db = self.db
        row = db['test'][0]
        assert_is_instance(row, SQLRow)

        row['a'] = 1
        row['b'] = 1
        assert_equal(db['test']['a'], [1, 11, 12, 13, 14,
                                       15, 16, 17, 18, 19])
        assert_equal(db['test']['b'], [1, 21, 22, 23, 24,
                                       25, 26, 27, 28, 29])

        with pytest.raises(KeyError):
            row['c'] = 1
        with pytest.raises(KeyError):
            row[2] = 1
        with pytest.raises(KeyError):
            row[-3] = 1

    def test_row_contains(self):
        db = self.db
        row = db['test'][0]
        assert_is_instance(row, SQLRow)

        assert_true(10 in row)
        assert_true(20 in row)
        assert_false('c' in row)
        assert_false('a' in row)
        assert_false('b' in row)

    def test_row_repr(self):
        db = self.db
        row = db['test'][0]
        assert_is_instance(row, SQLRow)
        assert_equal(repr(row), "SQLRow 0 in table 'test' {'a': 10, 'b': 20}")


class Test_SQLTable:
    @property
    def db(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))
        return db

    def test_table_basic_properties(self):
        db = self.db
        table = db['test']
        assert_equal(table.name, 'test')
        assert_equal(table.db, db.db)
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, list(zip(np.arange(10, 20),
                                            np.arange(20, 30))))

    def test_table_select(self):
        db = self.db
        table = db['test']

        a = table.select()
        assert_equal(a, list(zip(np.arange(10, 20),
                                 np.arange(20, 30))))

        a = table.select(order='a')
        assert_equal(a, list(zip(np.arange(10, 20),
                                 np.arange(20, 30))))

        a = table.select(order='a', limit=2)
        assert_equal(a, [(10, 20), (11, 21)])

        a = table.select(order='a', limit=2, offset=2)
        assert_equal(a, [(12, 22), (13, 23)])

        a = table.select(order='a', where='a < 15')
        assert_equal(a, [(10, 20), (11, 21), (12, 22), (13, 23), (14, 24)])

        a = table.select(order='a', where='a < 15', limit=3)
        assert_equal(a, [(10, 20), (11, 21), (12, 22)])

        a = table.select(order='a', where='a < 15', limit=3, offset=2)
        assert_equal(a, [(12, 22), (13, 23), (14, 24)])

        a = table.select(columns=['a'], where='a < 15')
        assert_equal(a, [(10,), (11,), (12,), (13,), (14,)])

    def test_table_as_table(self):
        db = self.db
        table = db['test']

        a = table.as_table()
        assert_is_instance(a, Table)
        assert_equal(a.colnames, ['a', 'b'])
        assert_equal(a, Table(names=['a', 'b'], data=[np.arange(10, 20),
                                                      np.arange(20, 30)]))

    def test_table_as_table_empty(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        table = db['test']

        a = table.as_table()
        assert_is_instance(a, Table)
        assert_equal(a.colnames, [])
        assert_equal(a, Table())

    def test_table_len(self):
        db = self.db
        table = db['test']
        assert_equal(len(table), 10)

    def test_table_iter(self):
        db = self.db
        table = db['test']

        v = 10
        for i in table:
            assert_equal(i, (v, v + 10))
            v += 1

    def test_table_contains(self):
        db = self.db
        table = db['test']

        assert_false(10 in table)
        assert_false(20 in table)
        assert_false('c' in table)
        assert_true('a' in table)
        assert_true('b' in table)

    def test_table_repr(self):
        db = self.db
        table = db['test']
        i = hex(id(table))

        expect = "SQLTable 'test' in database ':memory:':"
        expect += f"(2 columns x 10 rows)\n"
        expect += '\n'.join(table.as_table().__repr__().split('\n')[1:])
        assert_is_instance(table, SQLTable)
        assert_equal(repr(table), expect)

    def test_table_add_column(self):
        db = self.db
        table = db['test']

        table.add_column('c', data=np.arange(10, 20))
        assert_equal(table.column_names, ['a', 'b', 'c'])
        assert_equal(table.values, list(zip(np.arange(10, 20),
                                            np.arange(20, 30),
                                            np.arange(10, 20))))

        table.add_column('d', data=np.arange(20, 30))
        assert_equal(table.column_names, ['a', 'b', 'c', 'd'])
        assert_equal(table.values, list(zip(np.arange(10, 20),
                                            np.arange(20, 30),
                                            np.arange(10, 20),
                                            np.arange(20, 30))))

    def test_table_get_column(self):
        db = self.db
        table = db['test']

        a = table.get_column('a')
        assert_is_instance(a, SQLColumn)
        assert_equal(a.values, np.arange(10, 20))

        a = table.get_column('b')
        assert_is_instance(a, SQLColumn)
        assert_equal(a.values, np.arange(20, 30))

    def test_table_set_column(self):
        db = self.db
        table = db['test']

        table.set_column('a', np.arange(5, 15))
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, list(zip(np.arange(5, 15),
                                            np.arange(20, 30))))

    def test_table_set_column_invalid(self):
        db = self.db
        table = db['test']

        with assert_raises(ValueError):
            table.set_column('a', np.arange(5, 16))

        with assert_raises(KeyError):
            table.set_column('c', np.arange(5, 15))

    def test_table_add_row(self):
        db = self.db
        table = db['test']

        table.add_rows({'a': -1, 'b': -1})
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(len(table), 11)
        assert_equal(table[-1].values, (-1, -1))

        table.add_rows({'a': -2, 'c': -2}, add_columns=True)
        assert_equal(table.column_names, ['a', 'b', 'c'])
        assert_equal(len(table), 12)
        assert_equal(table[-1].values, (-2, None, -2))

        table.add_rows({'a': -3, 'd': -3}, add_columns=False)
        assert_equal(table.column_names, ['a', 'b', 'c'])
        assert_equal(len(table), 13)
        assert_equal(table[-1].values, (-3, None, None))

        # defult add_columns must be false
        table.add_rows({'a': -4, 'b': -4, 'c': -4, 'd': -4})
        assert_equal(table.column_names, ['a', 'b', 'c'])
        assert_equal(len(table), 14)
        assert_equal(table[-1].values, (-4, -4, -4))

    def test_table_add_row_invalid(self):
        db = self.db
        table = db['test']

        with assert_raises(ValueError):
            table.add_rows([1, 2, 3, 4])

        with assert_raises(TypeError):
            table.add_rows(2)

    def test_table_get_row(self):
        db = self.db
        table = db['test']

        a = table.get_row(0)
        assert_is_instance(a, SQLRow)
        assert_equal(a.values, (10, 20))

        a = table.get_row(1)
        assert_is_instance(a, SQLRow)
        assert_equal(a.values, (11, 21))

    def test_table_set_row(self):
        db = self.db
        table = db['test']

        table.set_row(0, {'a': 5, 'b': 15})
        expect = np.transpose([np.arange(10, 20), np.arange(20, 30)])
        expect[0] = [5, 15]
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        expect[-1] = [-1, -1]
        table.set_row(-1, {'a': -1, 'b': -1})
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        expect[-1] = [5, 5]
        table.set_row(-1, [5, 5])
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

    def test_table_set_row_invalid(self):
        db = self.db
        table = db['test']

        with pytest.raises(IndexError):
            table.set_row(10, {'a': -1, 'b': -1})
        with pytest.raises(IndexError):
            table.set_row(-11, {'a': -1, 'b': -1})

        with pytest.raises(TypeError):
            table.set_row(0, 'a')

    def test_table_getitem_int(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        assert_equal(table[0].values, (10, 20))
        assert_equal(table[-1].values, (19, 29))

        with pytest.raises(IndexError):
            table[10]
        with pytest.raises(IndexError):
            table[-11]

    def test_table_getitem_str(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        assert_equal(table['a'].values, np.arange(10, 20))
        assert_equal(table['b'].values, np.arange(20, 30))

        with pytest.raises(KeyError):
            table['c']

    def test_table_getitem_tuple(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        assert_equal(table[('a',)].values, np.arange(10, 20))
        assert_is_instance(table[('a',)], SQLColumn)
        assert_equal(table[(1,)].values, (11, 21))
        assert_is_instance(table[(1,)], SQLRow)

        with pytest.raises(KeyError):
            table[('c')]
        with pytest.raises(IndexError):
            table[(11,)]

    def test_table_getitem_tuple_rowcol(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        assert_equal(table['a', 0], 10)
        assert_equal(table['a', 1], 11)
        assert_equal(table['b', 0], 20)
        assert_equal(table['b', 1], 21)

        assert_equal(table[0, 'a'], 10)
        assert_equal(table[1, 'a'], 11)
        assert_equal(table[0, 'b'], 20)
        assert_equal(table[1, 'b'], 21)

        assert_equal(table['a', [0, 1, 2]], [10, 11, 12])
        assert_equal(table['b', [0, 1, 2]], [20, 21, 22])
        assert_equal(table[[0, 1, 2], 'b'], [20, 21, 22])
        assert_equal(table[[0, 1, 2], 'a'], [10, 11, 12])

        assert_equal(table['a', 2:5], [12, 13, 14])
        assert_equal(table['b', 2:5], [22, 23, 24])
        assert_equal(table[2:5, 'b'], [22, 23, 24])
        assert_equal(table[2:5, 'a'], [12, 13, 14])

        with pytest.raises(KeyError):
            table['c', 0]
        with pytest.raises(IndexError):
            table['a', 11]

        with pytest.raises(KeyError):
            table[0, 0]
        with pytest.raises(KeyError):
            table['b', 'a']
        with pytest.raises(KeyError):
            table[0, 1, 2]
        with pytest.raises(KeyError):
            table[0, 'a', 'b']

    def test_table_setitem_int(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        table[0] = {'a': 5, 'b': 15}
        expect = np.transpose([np.arange(10, 20), np.arange(20, 30)])
        expect[0] = [5, 15]
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        table[-1] = {'a': -1, 'b': -1}
        expect[-1] = [-1, -1]
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        with pytest.raises(IndexError):
            table[10] = {'a': -1, 'b': -1}
        with pytest.raises(IndexError):
            table[-11] = {'a': -1, 'b': -1}

    def test_table_setitem_str(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        table['a'] = np.arange(40, 50)
        expect = np.transpose([np.arange(40, 50), np.arange(20, 30)])
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        table['b'] = np.arange(10, 20)
        expect = np.transpose([np.arange(40, 50), np.arange(10, 20)])
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        with pytest.raises(KeyError):
            table['c'] = np.arange(10, 20)

    def test_table_setitem_tuple(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)

        table[('a',)] = np.arange(40, 50)
        expect = np.transpose([np.arange(40, 50), np.arange(20, 30)])
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        table[(1,)] = {'a': -1, 'b': -1}
        expect[1] = [-1, -1]
        assert_equal(table.column_names, ['a', 'b'])
        assert_equal(table.values, expect)

        with pytest.raises(KeyError):
            table[('c',)] = np.arange(10, 20)
        with pytest.raises(IndexError):
            table[(11,)] = np.arange(10, 20)

    def test_table_setitem_tuple_multiple(self):
        db = self.db
        table = db['test']
        assert_is_instance(table, SQLTable)
        expect = np.transpose([np.arange(10, 20), np.arange(20, 30)])

        table[('a', 1)] = 57
        expect[1, 0] = 57
        table['b', -1] = 32
        expect[-1, 1] = 32
        table[0, 'a'] = -1
        expect[0, 0] = -1
        table[5, 'b'] = 99
        expect[5, 1] = 99
        table['a', 3:6] = -999
        expect[3:6, 0] = -999
        table['b', [2, 7]] = -888
        expect[[2, 7], 1] = -888
        assert_equal(table.values, expect)

        with pytest.raises(KeyError):
            table[('c',)] = np.arange(10, 20)
        with pytest.raises(IndexError):
            table[(11,)] = np.arange(10, 20)
        with pytest.raises(KeyError):
            table['a', 'c'] = None
        with pytest.raises(KeyError):
            table[2:5] = 2
        with pytest.raises(KeyError):
            table[1, 2, 3] = 3

    def test_table_indexof(self):
        db = self.db
        table = db['test']
        assert_equal(table.index_of({'a': 15}), 5)
        assert_equal(table.index_of({'a': 50}), [])
        assert_equal(table.index_of('a < 13'), [0, 1, 2])


class Test_SQLColumn:
    @property
    def db(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))
        return db

    def test_column_basic_properties(self):
        db = self.db
        table = db['test']
        column = table['a']

        assert_equal(column.name, 'a')
        assert_equal(column.table, 'test')
        assert_equal(column.values, np.arange(10, 20))

    def test_column_len(self):
        db = self.db
        table = db['test']
        column = table['a']
        assert_equal(len(column), 10)

    def test_column_repr(self):
        db = self.db
        table = db['test']
        column = table['a']
        assert_equal(repr(column), "SQLColumn a in table 'test' (10 rows)")

    def test_column_contains(self):
        db = self.db
        table = db['test']
        column = table['a']
        assert_true(15 in column)
        assert_false(25 in column)

    def test_column_iter(self):
        db = self.db
        table = db['test']
        column = table['a']

        v = 10
        for i in column:
            assert_equal(i, v)
            v += 1

    def test_column_getitem_int(self):
        db = self.db
        table = db['test']
        column = table['a']

        assert_equal(column[0], 10)
        assert_equal(column[-1], 19)

        with pytest.raises(IndexError):
            column[10]
        with pytest.raises(IndexError):
            column[-11]

    def test_column_getitem_list(self):
        db = self.db
        table = db['test']
        column = table['a']

        assert_equal(column[[0, 1]], [10, 11])
        assert_equal(column[[-2, -1]], [18, 19])

        with pytest.raises(IndexError):
            column[[10, 11]]
        with pytest.raises(IndexError):
            column[[-11, -12]]

    def test_column_getitem_slice(self):
        db = self.db
        table = db['test']
        column = table['a']

        assert_equal(column[:2], [10, 11])
        assert_equal(column[-2:], [18, 19])
        assert_equal(column[2:5], [12, 13, 14])
        assert_equal(column[::-1], [19, 18, 17, 16, 15, 14, 13, 12, 11, 10])

    def test_column_getitem_tuple(self):
        db = self.db
        table = db['test']
        column = table['a']
        with pytest.raises(IndexError):
            column[('a',)]
        with pytest.raises(IndexError):
            column[(1,)]
        with pytest.raises(IndexError):
            column[1, 2]

    def test_column_setitem_int(self):
        db = self.db
        table = db['test']
        column = table['a']

        column[0] = 5
        assert_equal(db.get_row('test', 0).values, [5, 20])

        column[-1] = -1
        assert_equal(db.get_row('test', -1).values, [-1, 29])

    def test_column_setitem_list_slice(self):
        db = self.db
        table = db['test']
        column = table['a']

        column[:] = -1
        assert_equal(db.get_column('test', 'a').values, [-1]*10)
        column[[2, 4]] = 2
        assert_equal(db.get_column('test', 'a').values, [-1, -1, 2, -1, 2,
                                                         -1, -1, -1, -1, -1])
    def test_column_setitem_invalid(self):
        db = self.db
        table = db['test']
        column = table['a']

        with pytest.raises(IndexError):
            column[10] = 10
        with pytest.raises(IndexError):
            column[-11] = 10
        with pytest.raises(IndexError):
            column[2, 4] = [10, 11]


class Test_SQLTableMapping:
    @property
    def table(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        db.add_table('mapping')
        db.add_column('mapping', 'keywords', ['key a', 'key-b'])
        db.add_column('mapping', 'columns', ['a', 'b'])

        map = SQLColumnMap(db, 'mapping', 'keywords', 'columns')
        return SQLTable(db, 'test', colmap=map)

    def test_table_select(self):
        table = self.table

        a = table.select()
        assert_equal(a, list(zip(np.arange(10, 20),
                                 np.arange(20, 30))))

        a = table.select(order='key a')
        assert_equal(a, list(zip(np.arange(10, 20),
                                 np.arange(20, 30))))

        a = table.select(order='key-b', limit=2)
        assert_equal(a, [(10, 20), (11, 21)])

        a = table.select(order='key a', limit=2, offset=2)
        assert_equal(a, [(12, 22), (13, 23)])

        a = table.select(order='key-b', where={'key a': 15})
        assert_equal(a, [(15, 25)])

    def test_table_column_names(self):
        table = self.table
        assert_equal(table.column_names, ['key a', 'key-b'])

    def test_table_getitem_str(self):
        table = self.table

        assert_equal(table['key a'].values, np.arange(10, 20))
        assert_equal(table['key-b'].values, np.arange(20, 30))

        with pytest.raises(KeyError):
            table['c']

    def test_table_getitem_tuple(self):
        table = self.table
        assert_equal(table[('key a',)].values, np.arange(10, 20))
        assert_is_instance(table[('key a',)], SQLColumn)
        assert_equal(table[(1,)].values, (11, 21))
        assert_is_instance(table[(1,)], SQLRow)

        with pytest.raises(KeyError):
            table[('c')]
        with pytest.raises(IndexError):
            table[(11,)]

    def test_table_getitem_tuple_rowcol(self):
        table = self.table
        assert_equal(table['key a', 0], 10)
        assert_equal(table['key a', 1], 11)
        assert_equal(table['key-b', 0], 20)
        assert_equal(table['key-b', 1], 21)

        assert_equal(table[0, 'key a'], 10)
        assert_equal(table[1, 'key a'], 11)
        assert_equal(table[0, 'key-b'], 20)
        assert_equal(table[1, 'key-b'], 21)

        assert_equal(table['key a', [0, 1, 2]], [10, 11, 12])
        assert_equal(table['key-b', [0, 1, 2]], [20, 21, 22])
        assert_equal(table[[0, 1, 2], 'key-b'], [20, 21, 22])
        assert_equal(table[[0, 1, 2], 'key a'], [10, 11, 12])

        assert_equal(table['key a', 2:5], [12, 13, 14])
        assert_equal(table['key-b', 2:5], [22, 23, 24])
        assert_equal(table[2:5, 'key-b'], [22, 23, 24])
        assert_equal(table[2:5, 'key a'], [12, 13, 14])

        with pytest.raises(KeyError):
            table['c', 0]
        with pytest.raises(IndexError):
            table['key a', 11]

        with pytest.raises(KeyError):
            table[0, 0]
        with pytest.raises(KeyError):
            table['key-b', 'key a']
        with pytest.raises(KeyError):
            table[0, 1, 2]
        with pytest.raises(KeyError):
            table[0, 'key a', 'key-b']

    def test_table_set_row(self):
        table = self.table

        table.set_row(0, {'key a': 5, 'key-b': 15})
        expect = np.transpose([np.arange(10, 20), np.arange(20, 30)])
        expect[0] = [5, 15]
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        expect[-1] = [-1, -1]
        table.set_row(-1, {'key a': -1, 'key-b': -1})
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        expect[-1] = [5, 5]
        table.set_row(-1, [5, 5])
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

    def test_table_add_row(self):
        table = self.table

        table.add_rows({'key a': -1, 'key-b': -1})
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(len(table), 11)
        assert_equal(table[-1].values, (-1, -1))

        table.add_rows({'key a': -2, 'key!c': -2}, add_columns=True)
        assert_equal(table.column_names, ['key a', 'key-b', 'key!c'])
        assert_equal(len(table), 12)
        assert_equal(table[-1].values, (-2, None, -2))

        table.add_rows({'key a': -3, 'key_d': -3}, add_columns=False)
        assert_equal(table.column_names, ['key a', 'key-b', 'key!c'])
        assert_equal(len(table), 13)
        assert_equal(table[-1].values, (-3, None, None))

        # defult add_columns must be false
        table.add_rows({'key a': -4, 'key-b': -4, 'key!c': -4, 'key_d': -4})
        assert_equal(table.column_names, ['key a', 'key-b', 'key!c'])
        assert_equal(len(table), 14)
        assert_equal(table[-1].values, (-4, -4, -4))

    def test_table_get_column(self):
        table = self.table

        a = table.get_column('key a')
        assert_is_instance(a, SQLColumn)
        assert_equal(a.values, np.arange(10, 20))
        assert_equal(a.name, 'a')

        a = table.get_column('key-b')
        assert_is_instance(a, SQLColumn)
        assert_equal(a.values, np.arange(20, 30))
        assert_equal(a.name, 'b')

    def test_table_set_column(self):
        table = self.table

        table.set_column('key a', np.arange(5, 15))
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, list(zip(np.arange(5, 15),
                                            np.arange(20, 30))))

    def test_table_set_column_invalid(self):
        table = self.table

        with assert_raises(ValueError):
            table.set_column('key a', np.arange(5, 16))

        with assert_raises(KeyError):
            table.set_column('key!c', np.arange(5, 15))

    def test_table_add_column(self):
        table = self.table

        table.add_column('key!c', data=np.arange(10, 20))
        assert_equal(table.column_names, ['key a', 'key-b', 'key!c'])
        assert_equal(table.values, list(zip(np.arange(10, 20),
                                            np.arange(20, 30),
                                            np.arange(10, 20))))

        table.add_column('key_d', data=np.arange(20, 30))
        assert_equal(table.column_names, ['key a', 'key-b', 'key!c', 'key_d'])
        assert_equal(table.values, list(zip(np.arange(10, 20),
                                            np.arange(20, 30),
                                            np.arange(10, 20),
                                            np.arange(20, 30))))

    def test_table_contains(self):
        table = self.table

        assert_false(10 in table)
        assert_false(20 in table)
        assert_false('key!c' in table)
        assert_true('key a' in table)
        assert_true('key-b' in table)

    def test_table_as_table_empty(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_table('mapping')
        db.add_column('mapping', 'keywords')
        db.add_column('mapping', 'columns')
        table = SQLTable(db, 'test',
                         SQLColumnMap(db, 'mapping', 'keywords', 'columns'))

        a = table.as_table()
        assert_is_instance(a, Table)
        assert_equal(a.colnames, [])
        assert_equal(a, Table())

    def test_table_as_table(self):
        table = self.table

        a = table.as_table()
        assert_is_instance(a, Table)
        assert_equal(a.colnames, ['key a', 'key-b'])
        assert_equal(a, Table(names=['key a', 'key-b'],
                              data=[np.arange(10, 20), np.arange(20, 30)]))

    def test_table_setitem_int(self):
        table = self.table

        table[0] = {'key a': 5, 'key-b': 15}
        expect = np.transpose([np.arange(10, 20), np.arange(20, 30)])
        expect[0] = [5, 15]
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        table[-1] = {'key a': -1, 'key-b': -1}
        expect[-1] = [-1, -1]
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        with pytest.raises(IndexError):
            table[10] = {'key a': -1, 'key-b': -1}
        with pytest.raises(IndexError):
            table[-11] = {'key a': -1, 'key-b': -1}

    def test_table_setitem_str(self):
        table = self.table

        table['key a'] = np.arange(40, 50)
        expect = np.transpose([np.arange(40, 50), np.arange(20, 30)])
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        table['key-b'] = np.arange(10, 20)
        expect = np.transpose([np.arange(40, 50), np.arange(10, 20)])
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        with pytest.raises(KeyError):
            table['c'] = np.arange(10, 20)

    def test_table_setitem_tuple(self):
        table = self.table

        table[('key a',)] = np.arange(40, 50)
        expect = np.transpose([np.arange(40, 50), np.arange(20, 30)])
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

        table[(1,)] = {'key a': -1, 'key-b': -1}
        expect[1] = [-1, -1]
        assert_equal(table.column_names, ['key a', 'key-b'])
        assert_equal(table.values, expect)

    def test_table_setitem_tuple_multiple(self):
        table = self.table
        expect = np.transpose([np.arange(10, 20), np.arange(20, 30)])

        table[('key a', 1)] = 57
        expect[1, 0] = 57
        table['key-b', -1] = 32
        expect[-1, 1] = 32
        table[0, 'key a'] = -1
        expect[0, 0] = -1
        table[5, 'key-b'] = 99
        expect[5, 1] = 99
        table['key a', 3:6] = -999
        expect[3:6, 0] = -999
        table['key-b', [2, 7]] = -888
        expect[[2, 7], 1] = -888
        assert_equal(table.values, expect)

    def test_table_indexof(self):
        table = self.table
        assert_equal(table.index_of({'key a': 15}), 5)
        assert_equal(table.index_of({'key a': 50}), [])
        with pytest.raises(TypeError):
            assert_equal(table.index_of('"key a" < 13'), [0, 1, 2])


class Test_SQLRowMapping:
    @property
    def table(self):
        db = SQLDatabase(':memory:')
        db.add_table('test')
        db.add_column('test', 'a', data=np.arange(10, 20))
        db.add_column('test', 'b', data=np.arange(20, 30))

        db.add_table('mapping')
        db.add_column('mapping', 'keywords', ['key a', 'key-b'])
        db.add_column('mapping', 'columns', ['a', 'b'])

        map = SQLColumnMap(db, 'mapping', 'keywords', 'columns')
        return SQLTable(db, 'test', colmap=map)

    def test_row_column_names(self):
        table = self.table
        r = table.get_row(5)
        assert_equal(r.column_names, ['key a', 'key-b'])
        assert_equal(r.keys, ['key a', 'key-b'])
        r = table[5]
        assert_equal(r.column_names, ['key a', 'key-b'])
        assert_equal(r.keys, ['key a', 'key-b'])

    def test_row_invalid_index(self):
        table = self.table

        with pytest.raises(IndexError):
            table[10]
        with pytest.raises(IndexError):
            table[-11]
        with pytest.raises(IndexError):
            table.get_row(10)
        with pytest.raises(IndexError):
            table.get_row(-11)

    def test_row_getitem(self):
        table = self.table
        r = table.get_row(5)
        assert_equal(r['key a'], 15)
        assert_equal(r['key-b'], 25)

        with pytest.raises(KeyError):
            r['c']

    def test_row_setitem(self):
        table = self.table
        r = table.get_row(5)
        r['key a'] = -1
        r['key-b'] = -2
        assert_equal(r.values, [-1, -2])
        assert_equal(table.values[5], [-1, -2])

    def test_row_as_dict(self):
        table = self.table
        r = table.get_row(5)
        assert_equal(r.as_dict(), {'key a': 15, 'key-b': 25})

    def test_row_items(self):
        table = self.table
        r = table.get_row(5)
        assert_equal(list(r.items), [('key a', 15), ('key-b', 25)])
